import warnings

import numpy as np
from scipy.optimize import linprog

from relsad.network.systems import PowerSystem, Transmission
from relsad.Time import Time
from relsad.utils import INF

np.set_printoptions(suppress=True, linewidth=np.nan)
warnings.filterwarnings("ignore")

# flake8: noqa: C901
def shed_energy(
    power_system: PowerSystem,
    dt: Time,
    alpha: float = 1e-4,
):
    """
    Sheds the unsupplied loads of the power system over the
    time period, dt, using a linear minimization
    problem solved with linear programming

    See :doc:`/theory/opt` for more details.

    Parameters
    ----------
    power_system : PowerSystem
        The power system that is under consideration
    dt : Time
        The current time step
    alpha : float
        Slack variable to cope with numerical noise

    Problem formulation
    --------------------

    minimize sum(P_shed_n)

    such that
        sum(P_shed_n) - sum(P_line_nl) + sum(P_gen_n) = sum(P_load_n) for all buses n

                     0 <= P_shed_n  <= P_load_n
        -P_line_nl_max <= P_line_nl <= P_line_nl_max
           P_gen_n_min <= P_gen_n   <= P_gen_n_max

    The problem is solved with the following modification to prevent numerical difficulties

    such that
        sum(P_shed_n) - sum(P_line_nl) + sum(P_gen_n) + alpha = sum(P_load_n)  for all buses n

                     0 <= P_shed_n  <= P_load_n
        -P_line_nl_max <= P_line_nl <= P_line_nl_max
           P_gen_n_min <= P_gen_n   <= P_gen_n_max
             alpha_min <= alpha     <= alpha_max

    Returns
    -------
    None

    """
    buses = list(power_system.buses)
    lines = [x for x in power_system.lines if x.connected]
    N_D = len(buses)
    N_L = len(lines)
    # Define cost function
    c = [x.get_cost() for x in buses] + [0] * N_L + [0] * N_D + [0]
    # Get loads
    p_b = [max(0, bus.pload) for bus in buses]  # Active bus load
    q_b = [max(0, bus.qload) for bus in buses]  # Reactive bus load
    # Building lhs contraint matrix
    A = _build_A_matrix(power_system)
    # Gather bounds
    p_bounds, q_bounds = _gather_bounds(power_system, alpha)
    if sum(p_b) > alpha:
        # Shed active loads
        shedded_active_bus_loads = _shed_active_loads(
            c=c,
            A=A,
            p_b=p_b,
            p_bounds=p_bounds,
            buses=buses,
            lines=lines,
            alpha=alpha,
        )
        if shedded_active_bus_loads is not None:
            # Shed active energy
            _shed_active_energy(
                buses=buses,
                shedded_active_bus_loads=shedded_active_bus_loads,
                alpha=alpha,
                dt=dt,
            )
    if sum(q_b) > alpha:
        # Shed reactive loads
        shedded_reactive_bus_loads = _shed_reactive_loads(
            c=c,
            A=A,
            q_b=q_b,
            q_bounds=q_bounds,
            buses=buses,
            lines=lines,
            alpha=alpha,
        )
        if shedded_reactive_bus_loads is not None:
            # Shed reactive energy
            _shed_reactive_energy(
                buses=buses,
                shedded_reactive_bus_loads=shedded_reactive_bus_loads,
                alpha=alpha,
                dt=dt,
            )


def _build_A_matrix(power_system: PowerSystem):

    """
    Builds the LHS constraint matrix

    Parameters
    ----------
    power_system : PowerSystem
        The power system that is under consideration

    Returns
    -------
    A : array
        LHS Constraint matrix

    """

    buses = list(power_system.buses)
    lines = [x for x in power_system.lines if x.connected]
    N_D = len(buses)
    N_L = len(lines)
    A = np.zeros((N_D, N_D + N_L + N_D + 1))
    for j, bus in enumerate(buses):
        # Load shed coefficients
        A[j, j] = 1  # mu_md
        # Line load coefficients
        for line in bus.connected_lines:
            if line.connected:
                index = lines.index(line)
                if bus == line.fbus:
                    A[j, N_D + index] = -1
                else:
                    A[j, N_D + index] = 1
        # Generation coefficients
        A[j, N_D + N_L + j] = 1  # lambda_md
        # Slack parameter
        A[j, -1] = 1
    return A


def _get_generation_bounds(power_system: PowerSystem):
    """
    Returns the generation units bounds

    Parameters
    ----------
    power_system : PowerSystem
        The power system that is under consideration

    Returns
    -------
    p_gen : list
        List of the active power bounds for the generation units
    q_gen : list
        List of the reactive power bounds for the generation units

    """
    p_gen = list()  # Active bus generation
    q_gen = list()  # Reactive bus generation
    for j, bus in enumerate(power_system.buses):
        # Generation bounds
        flag = False
        for child_network in power_system.child_network_list:
            if isinstance(child_network, Transmission):
                if bus == child_network.get_trafo_bus():
                    p_gen.append(INF)
                    q_gen.append(INF)
                    flag = True
        if flag is False:
            p_gen.append(max(0, bus.pprod))
            q_gen.append(max(0, bus.qprod))
    return p_gen, q_gen


def _gather_bounds(
    power_system: PowerSystem,
    alpha: float,
):
    """
    Returns the flow bounds between the components in the power system.

    Parameters
    ----------
    power_system : PowerSystem
        The power system that is under consideration
    alpha : float
        Slack variable to cope with numerical noise

    Returns
    -------
    p_bounds : list
        The active power bounds for a component
    q_bounds : list
        The reactive power bounds for a component
    """

    # Gather bounds
    buses = list(power_system.buses)
    lines = [x for x in power_system.lines if x.connected]
    N_D = len(buses)
    N_L = len(lines)
    # Get loads
    p_b = [max(0, bus.pload) for bus in buses]  # Active bus load
    q_b = [max(0, bus.qload) for bus in buses]  # Reactive bus load
    p_bounds = list()
    q_bounds = list()
    # Get generation bounds
    p_gen, q_gen = _get_generation_bounds(power_system)
    for n in range(N_D + N_L + N_D):
        if n < N_D:
            # Bus load
            p_bounds.append((0, p_b[n]))
            q_bounds.append((0, q_b[n]))
        elif N_D <= n < N_D + N_L:
            # Line flow
            line = lines[n - N_D]
            # Line flow in MW
            max_available_p_flow = line.get_line_load()[0] * line.s_ref
            max_available_q_flow = line.get_line_load()[1] * line.s_ref
            PL_p_max = min(line.capacity, abs(max_available_p_flow))
            PL_q_max = min(line.capacity, abs(max_available_q_flow))
            p_bounds.append((-PL_p_max, PL_p_max))
            q_bounds.append((-PL_q_max, PL_q_max))
        else:
            # Bus generation
            p_bounds.append((0, p_gen[n - (N_D + N_L)]))
            q_bounds.append((0, q_gen[n - (N_D + N_L)]))
    # alpha bounds
    p_bounds.append((-alpha, alpha))
    q_bounds.append((-alpha, alpha))
    return p_bounds, q_bounds


def _shed_active_loads(
    c: list,
    A: list,
    p_b: list,
    p_bounds: list,
    buses: list,
    lines: list,
    alpha: float,
):
    """
    Sheds active power load

    Parameters
    ----------
    c : list
        Coefficients of the objective function
    A : list
        LHS constraint matrix
    p_b : list
        RHS constraint vector
    p_bounds : list
        Variable boundaries
    buses : list
        List containing buses
    lines : list
        List containing lines
    alpha : float
        Slack variable to cope with numerical noise
    dt : Time
        The current time step

    Returns
    -------
    active_shedded_bus_loads: list
        The active shedded bus loads of the current
        time increment

    """
    p_res = linprog(
        c,
        A_eq=A,
        b_eq=p_b,
        bounds=p_bounds,
    )
    if not p_res.success:
        print("Active energy.shed was not successful, verify the results:")
        print("buses: ", buses)
        print("lines: ", lines)
        print("A: ", A)
        print("cost: ", c)
        print("p_b: ", p_b)
        print("p_bounds: ", p_bounds)
        p_res = linprog(
            c,
            A_eq=A,
            b_eq=p_b,
            bounds=p_bounds,
            options={
                "disp": True,
            },
        )
        print(p_res)
    if p_res.fun > 0:
        shedded_active_bus_loads = p_res.x
    else:
        shedded_active_bus_loads = None
    return shedded_active_bus_loads


def _shed_active_energy(
    buses: list,
    shedded_active_bus_loads: list,
    alpha: float,
    dt: Time,
):
    """
    Sheds active energy

    Parameters
    ----------
    buses : list
        List containing buses
    shedded_active_bus_loads : list
        The shedded active bus loads of the current
        time increment
    alpha : float
        Slack variable to cope with numerical noise
    dt : Time
        The current time step

    Returns
    -------
    None

    """
    for i, bus in enumerate(buses):
        bus.add_to_energy_shed_stack(
            shedded_active_bus_loads[i]
            if shedded_active_bus_loads[i] > alpha
            else 0,
            0,
            dt,
        )


def _shed_reactive_loads(
    c: list,
    A: list,
    q_b: list,
    q_bounds: list,
    buses: list,
    lines: list,
    alpha: float,
):
    """
    Sheds reactive power load

    Parameters
    ----------
    c : list
        Coefficients of the objective function
    A : list
        LHS constraint matrix
    q_b : list
        RHS constraint vector
    q_bounds : list
        Variable boundaries
    buses : list
        List containing buses
    lines : list
        List containing lines
    alpha : float
        Slack variable to cope with numerical noise
    dt : Time
        The current time step

    Returns
    -------
    shedded_reactive_bus_loads : list
        The shedded reactive bus loads of the current
        time increment

    """
    q_res = linprog(
        c,
        A_eq=A,
        b_eq=q_b,
        bounds=q_bounds,
    )
    if not q_res.success:
        print("Reactive energy.shed was not successful, verify the results:")
        print("buses: ", buses)
        print("lines: ", lines)
        print("A: ", A)
        print("cost: ", c)
        print("q_b: ", q_b)
        print("q_bounds: ", q_bounds)
        q_res = linprog(
            c,
            A_eq=A,
            b_eq=q_b,
            bounds=q_bounds,
            options={
                "disp": True,
            },
        )
        print(q_res)
    if q_res.fun > 0:
        shedded_reactive_bus_loads = q_res.x
    else:
        shedded_reactive_bus_loads = None
    return shedded_reactive_bus_loads


def _shed_reactive_energy(
    buses: list,
    shedded_reactive_bus_loads: list,
    alpha: float,
    dt: Time,
):
    """
    Sheds reactive energy

    Parameters
    ----------
    buses : list
        List containing buses
    shedded_reactive_bus_loads : list
        The shedded reactive bus loads of the current
        time increment
    alpha : float
        Slack variable to cope with numerical noise
    dt : Time
        The current time step

    Returns
    -------
    None

    """
    for i, bus in enumerate(buses):
        bus.add_to_energy_shed_stack(
            0,
            shedded_reactive_bus_loads[i]
            if shedded_reactive_bus_loads[i] > alpha
            else 0,
            dt,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['relsad',
 'relsad.StatDist',
 'relsad.energy',
 'relsad.examples.CINELDI',
 'relsad.examples.IEEE16_modified',
 'relsad.examples.IEEE33',
 'relsad.examples.IEEE69',
 'relsad.examples.RBTS2',
 'relsad.examples.RBTS6',
 'relsad.examples.TEST10',
 'relsad.examples.load',
 'relsad.examples.tutorial',
 'relsad.load',
 'relsad.loadflow.ac',
 'relsad.network',
 'relsad.network.components',
 'relsad.network.containers',
 'relsad.network.systems',
 'relsad.reliability.indices',
 'relsad.results.storage',
 'relsad.simulation',
 'relsad.simulation.monte_carlo',
 'relsad.simulation.sequence',
 'relsad.topology',
 'relsad.topology.ICT',
 'relsad.topology.load_flow',
 'relsad.utils',
 'relsad.visualization.plotting']

package_data = \
{'': ['*'], 'relsad.examples.load': ['data/*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0', 'pandas>=1.4.1,<2.0.0', 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'relsad',
    'version': '0.2.12',
    'description': 'A package that facilitates reliability investigations of power systems',
    'long_description': '\n|language badge| |code coverage badge| |contributors badge| |python version badge|\n|relsad version badge| |build badge| |docs badge| |license badge| |joss badge|\n\n######\nRELSAD\n######\n\n.. overview_start\n\n`RELSAD` -- RELiability tool for Smart and Active Distribution networks,\nis a Python-based reliability assessment tool that aims to function as\na foundation for reliability calculation of modern distribution systems.\nThe tool allows for Monte Carlo simulation based reliability analysis of modern\ndistribution networks, and sequential simulation of the network\nbehavior with user-defined loading and failure evolution to investigate the impact\nof the introduction of for instance ICT components.\n\n\nThe package supports user-selected time steps over a user-defined time period.\nIn the tool, active components such as microgrids, distributed generation,\nbatteries, and electrical vehicles are implemented.\nTo evaluate smart power systems, ICT (Information and Communication Technology)\ncomponents such as automated switches, sensors, and control systems\nfor the power grid are also implemented.\nIn addition to component implementation, in order to evaluate the reliability\nof such complex systems, the complexity, dependencies within a system,\nand interdependencies between systems and components are accounted for.\nFor now, only radial systems are supported.\n\nThe tool can be used in modern distribution network development to evaluate\nthe influence of active components on the network reliability. Relevant use cases\ninclude investigating how:\n\n1. The introduction of microgrids with active generation\n   affects the customers in the distribution network and vice versa\n2. Vehicle\\-to\\-grid strategies might mitigate load peaks and\n   improve the distribution network reliability\n3. The reliability of the ICT network impacts the\n   distribution network reliability\n\n.. overview_end\n\n============\nInstallation\n============\n\nSee https://relsad.readthedocs.io/en/latest/installation.html.\n\n========\nFeatures\n========\n\n- Monte Carlo simulation based reliability analysis of modern distribution networks\n- Sequential simulation of the network behavoir with user-defined loading and failure evolution\n\n============\nDependencies\n============\n\nThe package dependencies can be found in `pyproject.toml`.\n\n=====\nUsage\n=====\n\nExamples using well known test networks are included and presented in\nhttps://relsad.readthedocs.io/en/latest/usage/main.html.\n\n=============\nDocumentation\n=============\n\nThe official documentation is hosted on Read the Docs: https://relsad.readthedocs.io/en/latest/\n\n============\nContributors\n============\n\nWe welcome and recognize all contributions. You can see a list of current contributors in the contributors tab.\n\n\n====\nHelp\n====\n\nIf you have questions, feel free to contact the author.\n\n\n.. |contributors badge| image:: https://img.shields.io/github/contributors/stinefm/relsad\n   :target: https://github.com/stinefm/relsad/graphs/contributors\n\n.. |language badge| image:: https://img.shields.io/github/languages/top/stinefm/relsad\n   :target: https://www.python.org/\n\n.. |code coverage badge| image:: https://img.shields.io/codecov/c/github/stinefm/relsad\n   :target: https://app.codecov.io/github/stinefm/relsad\n\n.. |python version badge| image:: https://img.shields.io/pypi/pyversions/relsad\n\n.. |relsad version badge| image:: https://img.shields.io/pypi/v/relsad\n   :target: https://pypi.org/project/relsad/\n\n.. |build badge| image:: https://img.shields.io/github/workflow/status/stinefm/relsad/ci-cd\n   :target: https://github.com/stinefm/relsad/actions\n\n.. |docs badge| image:: https://readthedocs.org/projects/relsad/badge/?version=latest\n   :target: https://relsad.readthedocs.io/en/latest/\n\n.. |license badge| image:: https://img.shields.io/github/license/stinefm/relsad\n   :target: https://github.com/stinefm/relsad/blob/main/LICENSE\n\n.. |joss badge| image:: https://joss.theoj.org/papers/89b8a25755bb2641370bf83b70666e0a/status.svg\n   :target: https://joss.theoj.org/papers/89b8a25755bb2641370bf83b70666e0a\n',
    'author': 'Stine Fleischer Myhre',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/stinefm/relsad',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
=============
Network types
=============

.. currentmodule:: relsad.network.systems


"""

from .Distribution import Distribution
from .ICTNetwork import ICTNetwork
from .Microgrid import Microgrid
from .PowerNetwork import PowerNetwork
from .PowerSystem import PowerSystem
from .SubSystem import SubSystem
from .Transmission import Transmission

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from relsad.examples.tutorial.system import *

from relsad.simulation import Simulation


def callback(ps, prev_time, curr_time):
    dt = curr_time - prev_time
    if curr_time <= dt:
        ps.get_comp("L2").fail(dt=dt)
        ps.get_comp("L6").fail(dt=dt)
    elif Time(1.95, unit=dt.unit) < curr_time < Time(2.05, unit=dt.unit):
        ps.get_comp("L3").fail(dt=dt)


sim = Simulation(power_system=ps, random_seed=0)
sim.run_sequential(
    start_time=TimeStamp(
        year=2019,
        month=0,
        day=0,
        hour=0,
        minute=0,
        second=0,
    ),
    stop_time=TimeStamp(
        year=2019,
        month=0,
        day=0,
        hour=10,
        minute=0,
        second=0,
    ),
    time_step=Time(0.1, TimeUnit.HOUR),
    time_unit=TimeUnit.HOUR,
    callback=callback,
    save_dir="results",
)

path = os.path.join(
    "results",
    "sequence",
    "ps1",
    "ENS.csv",
)

df = pd.read_csv(path)
fig, ax = plt.subplots()
df.plot(
    x="HOUR",
    y="ps1",
    ax=ax,
)

fig.savefig(
    "ENS.png",
    dpi=600,
)

print(df.describe())

import shutil

shutil.rmtree("results")
os.remove("ENS.png")



import matplotlib.lines as mlines
import numpy as np

from relsad.Time import Time, TimeUnit

from .Bus import Bus
from .Component import Component
from .Line import Line
from .Switch import Switch


class Disconnector(Component, Switch):

    """
    Common base class for disconnectors

    ...

    Attributes
    ----------
    name : string
        Name of the disconnector
    coordinate : list
        The coordinate of the disconnector
    initial_state : bool
        The initial state of the disconnector (open or colsed)
    is_open : bool
        Tells if the switch is open (True) or closed (False)
    failed : bool
        True if the disconnector is in a failed state, False if not
    fail_rate : float
        The failure rate of the disconnector [no. of fails per year]
    outage_time : Time
        The outage time of the diconnector
    line : Line
        The line the disconnecor is connected to
    base_bus : Bus
        Wich bus the disconnector is closes to (for setting coordinates)
    intelligent_switch : IntelligentSwitch
        Returns the intelligent switch connected to the disconnector if any
    history : dict
        Dictonary attribute that stores the historic variables

    Methods
    ----------
    close()
        Closes the disconnector
    open()
        Opens the disconnector
    fail()
        Sets the disconnector to failed and opens the disconnector
    not_fail()
        Sets the doconnector to not failed and closes the disconnector
    initialize_history()
        Initializes the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    reset_status(save_falg)
        Resets and sets the status of the system parameters

    """

    ## Visual attributes
    color = "black"
    edgecolor = "black"
    marker = "o"
    size = 2**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=2,
        markersize=size,
        linestyle="None",
        color=color,
        markeredgecolor=edgecolor,
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        line: Line,
        bus: Bus,
        is_open: bool = False,
        fail_rate: float = 0.014,
        outage_time: Time = Time(1, TimeUnit.HOUR),
    ):

        # Verify input
        if line is None:
            raise Exception("Disconnector must be connected to a Line")
        if line.parent_network is not None:
            raise Exception(
                "Disconnector must be created before the line is connected to a network"
            )
        if bus is None:
            raise Exception("Disconnector must be connected to a Bus")
        if bus not in [line.fbus, line.tbus]:
            raise Exception(
                "Disconnector must be connected to a base bus connected "
                "to the line it is located on"
            )
        if fail_rate < 0:
            raise Exception("The failure rate must be positive")

        self.name = name
        self.initial_state = is_open
        self.is_open = is_open
        self.failed = False
        self.fail_rate = fail_rate
        self.outage_time = outage_time
        self.line = line

        ## Set coordinate
        self.base_bus = bus
        dx = line.tbus.coordinate[0] - line.fbus.coordinate[0]
        dy = line.tbus.coordinate[1] - line.fbus.coordinate[1]
        if bus == line.tbus:
            dx *= -1
            dy *= -1

        line.disconnectors.append(self)
        self.coordinate = [
            self.base_bus.coordinate[0] + dx / 4,
            self.base_bus.coordinate[1] + dy / 4,
        ]

        ## Communication
        self.intelligent_switch = None

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Disconnector(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Disconnector)
        return False

    def __hash__(self):
        return hash(self.name)

    def close(self):
        """
        Closes the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = False
        self.color = "black"
        if not self.line.connected:
            self.line.connect()

    def open(self):
        """
        Opens the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = True
        self.color = "white"
        if self.line.connected:
            self.line.disconnect()

    def fail(self):
        """
        Sets the disconnector to failed and opens the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = True
        self.open()

    def not_fail(self):
        """
        Sets the disconnector to not failed and closes the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = False
        self.close()

    def update_fail_status(self, dt: Time):
        """

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["is_open"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["is_open"][time] = self.is_open

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """

        Parameters
        ----------

        Returns
        ----------
        None

        """

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.not_fail()
        if save_flag:
            self.initialize_history()


if __name__ == "__main__":
    pass



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def weather_generation_data(
    path: str = os.path.join("data", "weather_data_rygge.csv"),
):

    # Generate temperatures for a whole year
    data = pd.read_csv(path, header=None, skiprows=1)

    temp_profiles = data[1].values.reshape(365, 24)
    wind_profiles = data[2].values.reshape(365, 24)
    solar_profiles = data[3].values.reshape(365, 24)

    return temp_profiles, wind_profiles, solar_profiles


def load_data(
    temp: np.array,
    path: str = os.path.join("data", "load_profiles_fasit.csv"),
):

    # Generating load profiles:

    data = pd.read_csv(
        path,
        sep=";",
        decimal=",",
        header=None,
    ).transpose()
    data = data.stack().str.replace(",", ".").unstack()

    house_A = data[10].values[4:-1].astype(float)
    house_B = data[11].values[4:-1].astype(float)

    farm_A = data[2].values[4:-1].astype(float)
    farm_B = data[3].values[4:-1].astype(float)

    industry2_A = data[26].values[4:-1].astype(float)
    industry2_B = data[27].values[4:-1].astype(float)

    trade_A = data[42].values[4:-1].astype(float)
    trade_B = data[43].values[4:-1].astype(float)

    office_A = data[50].values[4:-1].astype(float)
    office_B = data[51].values[4:-1].astype(float)

    load_house = np.zeros_like(temp)
    load_farm = np.zeros_like(temp)
    load_microgrid = np.zeros_like(temp)
    load_industry2 = np.zeros_like(temp)
    load_trade = np.zeros_like(temp)
    load_office = np.zeros_like(temp)

    for i in range(len(temp)):
        for j in range(len(temp[i])):
            if temp[i, j] > 20:
                temp[i, j] = 20

            load_house[i, j] = house_A[j] * temp[i, j] + house_B[j]
            load_farm[i, j] = farm_A[j] * temp[i, j] + farm_B[j]
            load_microgrid[i, j] = farm_A[j] * temp[i, j] + farm_B[j]
            load_industry2[i, j] = industry2_A[j] * temp[i, j] + industry2_B[j]
            load_trade[i, j] = trade_A[j] * temp[i, j] + trade_B[j]
            load_office[i, j] = office_A[j] * temp[i, j] + office_B[j]

    return (
        load_house.reshape((8760)) * 1e-3,
        load_farm.reshape((8760)) * 1e-3,
        load_microgrid.reshape((8760)) * 1e-3,
        load_industry2.reshape((8760)) * 1e-3,
        load_trade.reshape((8760)) * 1e-3,
        load_office.reshape((8760)) * 1e-3,
    )


def wind_power(
    wind: np.array,
):

    # Wind power

    # https://en.wind-turbine-models.com/turbines/1829-aeolia-windtech-d2cf-200

    N = 5
    alpha = 0.25
    H_0 = 2
    H = 50

    v = np.zeros_like(wind)
    P_out = np.zeros_like(wind)

    P_rated = np.array(
        [
            0,
            0,
            0,
            34.8,
            69.7,
            123.5,
            200.7,
            306.6,
            446.4,
            495.0,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            500,
            0,
        ]
    )

    for i in range(len(wind)):
        for j in range(len(wind[i])):
            v[i, j] = wind[i, j] * (H / H_0) ** (alpha)

    for day in range(len(v)):
        for hour in range(len(v[day])):

            if v[day, hour] <= 3 or v[day, hour] > 25:
                P_out[day, hour] = 0

            elif v[day, hour] > 10 and v[day, hour] <= 25:
                P_out[day, hour] = 500 * 10 ** (-3) * N

            else:
                v_a = int(v[day, hour])
                v_b = v_a + 1
                P_out[day, hour] = (
                    N
                    * (
                        P_rated[v_a - 1]
                        + (P_rated[v_b - 1] - P_rated[v_a - 1])
                        * ((v[day, hour] - v_a) / (v_b - v_a))
                    )
                    * 10 ** (-3)
                )

    return P_out.reshape((8760))


def pv_power(
    temp: np.array,
    irridation: np.array,
):

    modules = 5000 / 2

    # Maximum power point voltage [V]
    Vmpp = 31.2
    # Maximum power point current [A]
    Impp = 8.18

    # Open circuit voltage [V]
    Voc = 37.8
    # Short circuit current [A]
    Isc = 8.89

    # Maximum power point power [W]
    Pmpp = Vmpp * Impp

    # Nominal operating cell temperature [C]
    NOCT = 47.0

    # Area of the modul [m^2]
    area_modul = 1.6

    # Referance temp (25 C) [K]
    T_0 = 25 + 273.15

    # Standard irradiance [W/m^2 ]
    E_0 = 1000

    # Fill factor
    FF = Pmpp / (Voc * Isc)

    # Constant [A/(W/m^2)]
    K = Isc / E_0

    # diode saturation current [A/m^2]
    I_0 = 10 ** (-12) * area_modul

    # Constant term K/I0 (=10^6 m^2/W)
    k1 = K / I_0

    # Inverter efficiency
    n_inv = 0.9

    # Fill factor modul constant [Km^2]
    # Cff = Voc*Isc*T0*FF/(E0*ln(10^6*E0))
    # Equation 9
    Cff = Voc * Isc * T_0 * FF / (E_0 * np.log(10 ** (6) * E_0))

    E = irridation

    # Temperatur on the cell
    # Tcell = Temp + (NOCT-20)/800*Global_inn + 273.15 [K]
    T_cell = np.zeros_like(temp)

    # Output power for each modulin [Wh/h]
    # FF*Isc*global_inn(E)/E_0*Voc*ln(k1*E)/ln(k1*E0)*T0/T_cell
    # Equation 5
    P_out = np.zeros_like(temp)

    # Output power after the inverter for all modules [Wh/h]
    # Pout * n_inv * #modules
    P_out_mod_inv = np.zeros_like(temp)

    # Output power [Wh/h], equtaion 8
    # FF * Isc*Voc*T_0/(E_0*(ln(10^6*E_0)))*n_inv*#modules*E*ln(10^6*E)/T_cell
    # Cff * #modules*n_inv*(E*ln(10^6*E))/T_cell
    P_out_simp = np.zeros_like(temp)

    for i in range(len(temp)):
        for j in range(len(temp[i])):

            if irridation[i, j] <= 0.0:

                T_cell[i, j] = temp[i, j] + 273.15
            else:
                T_cell[i, j] = (
                    temp[i, j]
                    + ((NOCT - 20) / 800) * irridation[i, j]
                    + 273.15
                )

            if E[i, j] <= 0.0:
                P_out[i, j] = 0
                P_out_mod_inv[i, j] = 0
                P_out_simp[i, j] = 0

            else:
                P_out[i, j] = (
                    FF
                    * (Isc * (E[i, j] / E_0))
                    * (
                        Voc
                        * (np.log(k1 * E[i, j]) / np.log(k1 * E_0))
                        * (T_0 / T_cell[i, j])
                    )
                )
                P_out_mod_inv[i, j] = P_out[i, j] * n_inv * modules
                P_out_simp[i, j] = (
                    Cff
                    * modules
                    * n_inv
                    * ((E[i, j] * np.log(10 ** (6) * E[i, j])) / T_cell[i, j])
                )

    P_out_mod_inv *= 10 ** (-6)

    return P_out_mod_inv.reshape((8760))


if __name__ == "__main__":

    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    temp_profiles, wind_profiles, solar_profiles = weather_generation_data(
        path=os.path.join(
            os.pardir, os.pardir, "data", "weather_data_rygge.csv"
        ),
    )

    wind = wind_power(wind_profiles)
    PV = pv_power(temp_profiles, solar_profiles)

    fig1, ax1 = plt.subplots()
    ax1.plot(wind.flatten(), label="wind")
    ax1.plot(PV.flatten(), label="PV")
    ax1.legend()

    (
        load_house,
        load_farm,
        load_microgrid,
        load_industry2,
        load_trade,
        load_office,
    ) = load_data(
        temp=temp_profiles,
        path=os.path.join(
            os.pardir, os.pardir, "data", "load_profiles_fasit.csv"
        ),
    )

    fig2, ax2 = plt.subplots()
    ax2.plot(load_house.flatten(), label="house")
    ax2.plot(load_farm.flatten(), label="farm")
    ax2.plot(load_microgrid.flatten(), label="microgrid")
    ax2.plot(load_industry2.flatten(), label="industry")
    ax2.plot(load_trade.flatten(), label="trade")
    ax2.plot(load_office.flatten(), label="office")
    ax2.legend()

    fig3, ax3 = plt.subplots()
    ax3.plot(load_microgrid.flatten() * 40, label="load", zorder=3)
    ax3.plot(wind.flatten(), label="wind")
    ax3.plot(PV.flatten(), label="PV")
    ax3.legend()

    fig4, ax4 = plt.subplots()
    ax4.plot(load_house.flatten() * 400, label="load house", zorder=3)
    ax4.plot(load_farm.flatten() * 340, label="load farm", zorder=3)
    ax4.plot(load_industry2.flatten() * 5, label="load industry", zorder=3)
    ax4.plot(load_office.flatten() * 4, label="load office", zorder=3)
    ax4.plot(load_trade.flatten() * 7, label="load trade", zorder=3)

    ax4.legend()

    house = np.max(load_house) * 357  # * 300  # 400
    farm = np.max(load_farm) * 273  # * 250  # 340
    industry = np.max(load_industry2) * 8  # * 4  # 5
    office = np.max(load_office) * 3  # * 4
    trade = np.max(load_trade) * 2  # * 6  # 7
    print("Max load house:", house)
    print("Max load farm:", farm)
    print("Max load industry:", industry)
    print("Max load office:", office)
    print("Max load trade:", trade)
    print("Wind:", np.max(wind))
    print("PV:", np.max(PV))
    print("Sum:", np.sum(house + farm + industry + office + trade))
    print(sum(load_house))
    print(sum(load_farm))

    plt.show()



from abc import ABC, abstractmethod

from relsad.Time import Time


class Component(ABC):
    @abstractmethod
    def update_fail_status(self, dt: Time):
        pass

    @abstractmethod
    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        pass

    @abstractmethod
    def get_history(self, attribute: str):
        pass

    @abstractmethod
    def add_random_instance(self, random_gen):
        pass

    @abstractmethod
    def print_status(self):
        pass

    @abstractmethod
    def reset_status(self, save_flag: bool):
        pass



import numpy as np

from relsad.network.systems.PowerNetwork import PowerNetwork
from relsad.topology.load_flow.bfs import (
    configure_bfs_load_flow_setup,
    is_cyclic,
)


def run_bfs_load_flow(network: PowerNetwork, maxit: int = 5):
    """
    Solves the load flow with a specified number of iterations
    The two first septs are to set up additions topology information
    and to build the main structure
    Next, it is switched between forward sweeps (Voltage updates)
    and backward sweeps(load update and loss calcuation)

    See :doc:`/theory/bfs` for more details.

    Parameters
    ----------
    network : PowerNetwork
        The analyzed network
    maxit : int
        The number of iterations

    Returns
    -------
    network.buses : list
        List of network buses
    """
    # Check if network is radial
    if is_cyclic(network) is True:
        raise Exception("The network is not radial, cannot run BFS")

    # Configure topology list
    topology_list, network.buses = configure_bfs_load_flow_setup(
        network.buses, network.lines
    )

    # Run Backward Forward Sweep
    for _ in range(maxit):
        accumulate_load(topology_list)
        update_voltage(topology_list)
    return network.buses


def accumulate_load(topology_list):
    """
    Calculates the accumulated downstream active and
    reactive load at all buses and calculates the active
    and reactive losses of lines and make an accumulated
    equivalent load at the buses
    The function returns the accumulated loads and the
    accumulated power losses from a branch in the system

    Parameters
    ----------
    topology_list : list
        List containing the system topology

    Returns
    -------
    p_load : float
        Active accumulated power load of a branch
    q_load : float
        Reactive accumulated power load of a branch
    p_loss : float
        Active accumulated power loss of a branch
    q_loss : float
        Reactive accumulated power loss of a branch
    """
    p_load = 0.0
    q_load = 0.0
    p_loss = 0.0
    q_loss = 0.0
    for branch in reversed(topology_list):  # Start on last node
        if len(branch) > 1:
            for child_branch in branch[1:]:  # Do for all child branches
                (
                    p_load_child,
                    q_load_child,
                    p_loss_child,
                    q_loss_child,
                ) = accumulate_load(child_branch)
                # Add accumulated powers and losses in a branch to
                # the node where the brancing accurs.
                p_load += p_load_child
                q_load += q_load_child
                p_loss += p_loss_child
                q_loss += q_loss_child
        parent_bus = branch[0]
        # Add local loads
        p_load_local, q_load_local = get_load(parent_bus)
        p_load += p_load_local
        q_load += q_load_local
        # Add accumulated descriptions to the branching node
        parent_bus.p_load_downstream = p_load
        parent_bus.q_load_downstream = q_load

        # Add line loss
        if parent_bus.toline:
            to_line = parent_bus.toline
            if to_line.connected:
                # Find the flow to the downstream bus
                pto = parent_bus.p_load_downstream + p_loss
                qto = parent_bus.q_load_downstream + q_loss
                # Estimate the losses of the branch
                to_line.ploss = (
                    to_line.r_pu
                    * (pto**2 + qto**2)
                    / parent_bus.vomag**2
                )
                to_line.qloss = (
                    to_line.x_pu
                    * (pto**2 + qto**2)
                    / parent_bus.vomag**2
                )
                p_loss += to_line.ploss
                q_loss += to_line.qloss
                # Add the losses to the downstream bus
                parent_bus.p_loss_downstream = p_loss
                parent_bus.q_loss_downstream = q_loss

    # Return the accumulated loads and losses from the current branch
    return p_load, q_load, p_loss, q_loss


#
# Function for calculating the voltages and sensitivities in the single phase case
#
def calc_bus_voltage_sensitivity_single_phase(fbus, tbus, tline, branch: list):
    """
    Calculate the node voltages and sensitivities in the single phase case

    Parameters
    ----------
    fbus : Bus
        The sending Bus element
    tbus : Bus
        The recieving Bus element
    tline : Line
        The recieving Line element
    branch : list
        List of branches in the system

    Returns
    -------
    None

    """

    vk2 = fbus.vomag**2
    # Find the accumulated loads and losses flowing on the branch
    tpload = branch[0].p_load_downstream + branch[0].p_loss_downstream
    tqload = branch[0].q_load_downstream + branch[0].q_loss_downstream
    # Voltage calculation
    term2 = 2 * (tpload * tline.r_pu + tqload * tline.x_pu)
    term3 = (
        (tpload**2 + tqload**2)
        * (tline.r_pu**2 + tline.x_pu**2)
        / fbus.vomag**2
    )
    # Update the bus voltage magnitude on the down-stream bus
    tbus.vomag = np.sqrt(vk2 - term2 + term3)

    # Voltage angle calculation
    busvoltreal = (
        fbus.vomag - (tpload * tline.r_pu + tqload * tline.x_pu) / fbus.vomag
    )
    busvoltimag = (tqload * tline.r_pu - tpload * tline.x_pu) / fbus.vomag
    tbus.voang = fbus.voang + np.arctan2(
        busvoltimag, busvoltreal
    )  # Update voltage angles


#
# Update the voltage profile starting on the top node
#
def update_voltage(topology_list):
    """
    Update the voltage profile based on the accumulated load on each bus

    Parameters
    ----------
    topology_list : list
        List containing the system topology

    Returns
    -------
    None
    """
    for branch in topology_list:
        if branch[0].toline:
            tline = branch[0].toline
            fbus = tline.fbus
            tbus = tline.tbus
            # Update voltages and sensitivities Single Phase
            calc_bus_voltage_sensitivity_single_phase(
                fbus, tbus, tline, branch
            )
        if len(branch) > 1:
            for bus in branch[1:]:  # Update voltages along the branches
                update_voltage(bus)


#
# Calculates the load for the actual volage at the bus
#
def get_load(bus):
    """
    Calculates the net voltage corrected load at the bus - currently a simple ZIP model is applied.

    Parameters
    ----------
    bus : Bus
        A Bus element

    Returns
    -------
    p_load_corr : float
        Corrected active power load
    q_load_corr : float
        Corrected reactive power load

    """
    # load - production [PU]
    relative_pload = bus.pload_pu - bus.pprod_pu
    relative_qload = bus.qload_pu - bus.qprod_pu

    p_load_corr = relative_pload * (
        bus.ZIP[0] * bus.vomag**2 + bus.ZIP[1] * bus.vomag + bus.ZIP[2]
    )
    q_load_corr = relative_qload * (
        bus.ZIP[0] * bus.vomag**2 + bus.ZIP[1] * bus.vomag + bus.ZIP[2]
    )
    return p_load_corr, q_load_corr



from enum import Enum

import matplotlib.lines as mlines
import numpy as np

from relsad.Time import Time, TimeUnit
from relsad.utils import convert_yearly_fail_rate, random_choice

from .Component import Component
from .Disconnector import Disconnector
from .ICTNode import ICTNode


class IntelligentSwitchState(Enum):
    """
    Intelligent switch state

    Attributes
    ----------
    OK : int
        The intelligent switch is up and running
    FAILED : int
        The intelligent switch has failed
    REPAIR : int
        The intelligent switch is being repaired
    """

    OK = 1
    FAILED = 2
    REPAIR = 3


class IntelligentSwitch(Component):

    """
    Common class for batteries
    ...

    Attributes
    ----------
    name : string
        Name of the intelligent switch
    disconnector : Disconnector
        The disconnector the intelligent switch is connected to
    ict_node : ICTNode
        The ICT node connected to the intelligent switch
    fail_rate_per_year : float
        The failure rate per year for the intelligent switch
    manual_repair_time : Time
        The time it takes to manually repair the intelligent switch
    state : IntelligentSwitchState
        Which state the intelligent switch is in
    remaining_repair_time : Time
        The remaining repair time of the intelligent switch
    history : dict
        Dictonary attribute that stores the historic variables
    monte_carlo_history : dict
        Dictonary attribute that stores the historic variables from the Monte Carlo simulation

    Methods
    ----------
    fail()
        Sets the intelligent switch state to FAILED
    not_fail()
        Sets the intelligent switch state to OK
    draw_fail_status(dt)
        Draws the state of the intelligent switch for a given time step
    draw_status(prob)
        Sets the state of the intelligent switch based on the probability of the state being FAILED
    get_open_repair_time(dt)
        Returns the time it takes to open the intelligent switch when it must be repaired
    get_open_time(dt)
        Returns the time it takes to open the intelligent switch based on the status
    open()
        Opens the disconnector
    repair_close(dt)
        Sets the remaining repair time of the intelligent switch
        Closes the disconnector
        Sets the state of the intelligent switch to repair
    close(dt)
        Closes the disconnector if the state of the intelligent switch is OK
    update_fail_status(dt)
        Updates the fail status of the intelligent switch
        If the state of the intelligent switch is REPAIR, the remaining repair time is calculated
        If the state of the intelligent switch is OK, the state of the intelligent switch is drawn
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random instance
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets the status of the intelligent switch
    initialize_history()
        Initializes the history variables
    """

    ## Visual attributes
    color = "seagreen"
    marker = "x"
    size = 2**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=1,
        markersize=size,
        color=color,
        linestyle="None",
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        disconnector: Disconnector,
        ict_node: ICTNode = None,
        fail_rate_per_year: float = 0.03,
        manual_repair_time: Time = Time(2, TimeUnit.HOUR),
        state: IntelligentSwitchState = IntelligentSwitchState.OK,
    ):

        # Verify input
        if disconnector is None:
            raise Exception(
                "IntelligentSwitch must be connected to a Disconnector"
            )
        if fail_rate_per_year < 0:
            raise Exception("The failure rate per year must be positive")

        self.name = name
        self.disconnector = disconnector
        disconnector.intelligent_switch = self
        self.ict_node = ict_node
        if ict_node is not None:
            ict_node.coordinate = self.disconnector.coordinate
        self.fail_rate_per_year = fail_rate_per_year
        self.remaining_repair_time = Time(0)
        self.manual_repair_time = manual_repair_time
        self.state = state

        ## History
        self.history = {}
        self.monte_carlo_history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"IntelligentSwitch(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, IntelligentSwitch
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def fail(self):
        """
        Sets the intelligent switch state to FAILED

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = IntelligentSwitchState.FAILED

    def not_fail(self):
        """
        Sets the intelligent switch state to OK

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = IntelligentSwitchState.OK

    def draw_fail_status(self, dt: Time):
        """
        Draws the state of the intelligent switch for a given time step

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        p_fail = convert_yearly_fail_rate(self.fail_rate_per_year, dt)
        self.draw_status(p_fail)

    def draw_status(self, prob):
        """
        Sets the state of the intelligent switch based on the probability of the state being FAILED

        Parameters
        ----------
        prob : float
            The probability that the intelligent switch state is FAILED

        Returns
        ----------
        None

        """
        if random_choice(self.ps_random, prob):
            self.fail()
        else:
            self.not_fail()

    def get_open_repair_time(self, dt: Time):
        """
        Returns the time it takes to open the intelligent switch
        when it must be repaired

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        open_repair_time: Time
            The time it takes to open the intelligent switch
            when it must be repaired

        """
        self.remaining_repair_time = self.manual_repair_time
        self.state = IntelligentSwitchState.REPAIR
        open_repair_time = (
            self.disconnector.line.parent_network.controller.manual_sectioning_time
        )
        return open_repair_time

    def get_open_time(self, dt: Time):
        """
        Returns the time it takes to open the intelligent switch
        based on the status

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        open_time: Time
            The time it takes to open the intelligent switch
            based on the status

        """
        open_time = None
        if self.state == IntelligentSwitchState.REPAIR:
            open_time = Time(0)
        else:
            if self.state == IntelligentSwitchState.OK:
                open_time = Time(0)
            elif self.state == IntelligentSwitchState.FAILED:
                open_time = self.get_open_repair_time(dt)
        return open_time

    def open(self):
        """
        Opens the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.disconnector.open()

    def repair_close(self, dt: Time):
        """
        Sets the remaining repair time of the intelligent switch
        Closes the disconnector
        Sets the state of the intelligent switch to repair

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        # Repair is started
        self.remaining_repair_time = self.manual_repair_time
        # Repair crew is assumed to be present repairing the line,
        # they close the disconnector manually
        self.disconnector.close()
        self.state = IntelligentSwitchState.REPAIR

    def close(self, dt: Time):
        """
        Closes the disconnector if the state of the intelligent switch is OK


        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if not self.state == IntelligentSwitchState.REPAIR:
            if self.state == IntelligentSwitchState.OK:
                self.disconnector.close()
            elif self.state == IntelligentSwitchState.FAILED:
                self.repair_close(dt)

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the intelligent switch
        If the state of the intelligent switch is REPAIR, the remaining repair time is calculated
        If the state of the intelligent switch is OK, the state of the intelligent switch is drawn

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.state == IntelligentSwitchState.REPAIR:
            self.remaining_repair_time -= dt
            if self.remaining_repair_time <= Time(0):
                self.not_fail()
        elif self.state == IntelligentSwitchState.OK:
            self.draw_fail_status(dt)

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            The current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["remaining_repair_time"][
                time
            ] = self.remaining_repair_time.get_unit_quantity(curr_time.unit)
            self.history["state"][time] = self.state.value

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute
        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random instance

        Parameters
        ----------
        random_gen : np.random.default_rng()
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass

    def reset_status(self, save_flag: bool):
        """
        Resets the status of the intelligent switch

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.state = IntelligentSwitchState.OK
        self.remaining_repair_time = Time(0)
        if save_flag:
            self.initialize_history()

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["remaining_repair_time"] = {}
        self.history["state"] = {}
